;;; -*- Syntax: common-lisp; Mode: LISP; Package: (parser :use common-lisp parser-util); Base: 10 -*- 

;;  This file contains the functions that lookup values from the lexicon

(in-package 'parser :use '(parser-util lisp))


;; Parts of Speech
;; The following function rules define the parts of speech used in the grammar
;; in terms of dictionary predicates

(defun-nondeterministic noun (phrase xlist-stack)
  (mvlet (((term rem1 stack1)
	   (terminal phrase xlist-stack)))
	 (mvlet (((noun-val agmt) 
		  (noun-form term)))
		(returned-values `(noun ,noun-val) agmt rem1 stack1))))

(defun-nondeterministic noun-form (term)
  (let ((features (get term 'noun)))
    (if features
	(let ((noun-val (car features))
	      (agmt (cadr features)))
	  (returned-values noun-val (list agmt)))
	(fail))))

	

(defun-nondeterministic det (phrase xlist-stack)
  (either (mvlet (((non-term stack) 
		   (virtual xlist-stack)))
		 (if (eq non-term 'det)
		     (returned-values phrase stack)
		     (fail)))
	  (mvlet (((term rem1 stack1)
		   (terminal phrase xlist-stack)))
		 (mvlet (((parsed-det num def)
			  (det0 term)))
			(returned-values parsed-det num def rem1 stack1)))
	  (returned-values 'generic '_ 'generic phrase xlist-stack)))


;; search for a det value in the assoc list, det's should be saved in the list
;; with a (2d-put! 'the 'det '(NO def)) (2d-put! 'a 'det '(sin indef))
(defun-nondeterministic det0 (term)
  (let ((features (get term 'det)))
    (if (not features)
	(fail)
	(returned-values (list 'det term) (car features) (cadr features)))))

(defun-nondeterministic adjs (phrase xlist-stack)
  (either (mvlet (((parsed-adj rem stack)
		   (pre-adj phrase xlist-stack)))
		 (mvlet (((rem-adjs rem2 stack2) 
			  (adjs rem stack)))
			(returned-values (append (list parsed-adj) rem-adjs)
				rem2 stack2)))
	  (returned-values '() phrase xlist-stack)))



(defun-nondeterministic pre-adj (phrase xlist-stack)
  (either (mvlet (((parsed-adj type rem stack)
		   (adj phrase xlist-stack)))
		 (returned-values parsed-adj rem stack))
	  (mvlet (((parsed-adj rem stack)
		   (sup-phrase phrase xlist-stack)))
		 (returned-values parsed-adj rem stack))))



(defun-nondeterministic sup-phrase (phrase xlist-stack)
  (either (mvlet (((parsed-adj rem1 stack1)
		   (sup-adj phrase xlist-stack)))
	      (returned-values (list 'sup (list 'most parsed-adj)) 
			       rem1 stack1))
	  (mvlet (((parsed-adv rem1 stack1)
		   (sup-adv phrase xlist-stack)))
	      (mvlet (((parsed-adj type rem2 stack2)
		       (adj rem1 stack1)))
		  (if (spec-eq? type 'quant)
		      (returned-values (list 'sup parsed-adj) rem2 stack2) ;; *BUG ret Op?
		      (fail))))))
	    
;; define comp-phrase later..

(defun-nondeterministic adj (phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-adj type) 
		  (adj0 term)))
		(returned-values parsed-adj type rem stack))))

(defun-nondeterministic adj0 (term)
  (either (let ((type (get term 'adj)))
    (if (not type)
	(fail)   
	(returned-values (list 'adj term) type)))
	  (fail)))

(defun-nondeterministic prep (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-prep)
		  (prep0 term)))
		(returned-values parsed-prep rem stack))))

(defun-nondeterministic prep0 (term)
  (let ((feature (get term 'prep)))
    (if (not feature)
	(fail)
	(returned-values (list 'prep term)))))

(defun-nondeterministic rel-adj (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-abs-adj)
		  (rel-adj0 term)))
		(returned-values parsed-abs-adj rem stack))))

(defun-nondeterministic rel-adj0 (term)
  (let ((abs-adj (get term 'rel-adj)))
    (if (not abs-adj)
	(fail)
	(returned-values (list 'adj abs-adj)))))

(defun-nondeterministic sup-adj (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-abs-adj)
		  (sup-adj0 term)))
		(returned-values parsed-abs-adj rem stack))))

(defun-nondeterministic sup-adj0 (term)
  (let ((abs-adj (get term 'sup-adj)))
    (if (not abs-adj)
	(fail)  
	(returned-values (list 'adj abs-adj)))))

(defun-nondeterministic comp-adv (phrase xlist-stack)
  (either (mvlet (((term rem stack) 
		   (terminal phrase xlist-stack)))
		 (if (eq term 'less);; ** can probably extend this to
		     (returned-values term rem stack);; elminate the non-determinism
		     (fail)));; but this is consistent w/prolog!  
	  (mvlet (((term rem stack) 
		   (terminal phrase xlist-stack)))
		 (if (eq term 'more)
		     (returned-values term rem stack)
		     (fail)))))


(defun-nondeterministic sup-adv (phrase xlist-stack)
  (either (mvlet (((term rem stack)
		   (terminal phrase xlist-stack)))
		 (if (eq term 'least);; ** can probably extend this to
		     (returned-values term rem stack);; elminate the non-determinism
		     (fail)));; but this is consistent w/prolog! 
	  (mvlet (((term rem stack)
		   (terminal phrase xlist-stack)))
		 (if (eq term 'most)
		     (returned-values term rem stack)
		     (fail)))))



(defun-nondeterministic sup-op (adv phrase xlist-stack)
  (either (if (spec-eq? adv 'least)
	      (returned-values (list '(not adv)) phrase xlist-stack)
	      (fail))
	  (if (spec-eq? adv 'most)
	      (returned-values (list '(not adv)) phrase xlist-stack)
	      (fail))))
  
		       
(defun-nondeterministic rel-pron (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((case1) 
		  (rel-pron0 term)))
		(returned-values case1 rem stack))))

(defun-nondeterministic rel-pron0 (term)
  (let ((case1 (get term 'rel-pron)))
    (if (not case1)
	(fail)
	(returned-values case1))))

(defun-nondeterministic verb-form (agmt phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (mvlet (((verb-agmt root verb-to tense type) 
		  (verb-form0 term agmt)))
		(if (and (match? tense '(_ fin))
			 (match? agmt verb-agmt))
		    (returned-values root verb-to tense type rem stack)
		    (fail))))) 
                           ;; null is Role? maybe!
                           ;;;  possible bug in paper Role is deleted!!


;; verbs saved in the alist will have the following form:
;; (2d-put! verb-value synt-cat features)  where synt-cat is 'verb
;; and features is a list of (root tense agmt type).  Furthermore, tense, agmt
;; and type are all lists of symbols.
;; For example: (2d-put! 'am 'verb (be (pres fin) (1 sin) (aux)))

(defun-nondeterministic verb-form0 (term agmt)
  (let ((features (get term 'verb)))
    (if features
	(let ((root (car features))
	      (tense (cadr features))
	      (verb-agmt (caddr features))
	      (type (cadddr features)))
	  (returned-values verb-agmt root `(verb ,term) tense type))
    (fail))))






(defun-nondeterministic noun-form (term)
  (let ((features (get term 'noun)))
    (if features
	(let ((noun-val (car features))
	      (agmt (cadr features)))
	  (returned-values noun-val (list agmt)))
	(fail))))

;; (defun-nondeterministic verb-form00 (term agmt)
;;   (let ((features (get term 'verb)))
;;     (if features
;; 	(let ((root (car features))
;; 	      (tense (cadr features))
;; 	      (verb-agmt (caddr features))
;; 	      (type (cadddr features)))
;; 	  ;(either)
;	  (if (and (match? tense '(inf)))
;; 	;		   (spec-eq? root term))
;; 		;      (returned-values `(verb ,term) tense type)
;; 		  ;    (fail)
;; 		  (if (and (match? tense '(pres fin))
;; 			   (match? agmt verb-agmt)
;; 			   (spec-eq? root term))
;; 		      (returned-values `(verb ,term) tense type))
;; 		      (fail))
;; 	;	      (fail)
;; 		;      (if (match? tense '(past _)) )
;; 	;		  (returned-values `(verb ,root) tense type)
;			  (fail)
;; 	  (fail))))
;;



(defun-nondeterministic name (phrase xlist-stack)
  (mvlet (((rem1 stack1) 
	   (opt-the phrase xlist-stack)))
	 (mvlet (((term rem2 stack2)
		  (terminal rem1 stack1)))
		(name0 term)
		(returned-values term rem2 stack2))))

;; this function elminates the function name-template which is supposed to call
;; diff functions (clauses) to distinguish between the diff databases (cities, 
;; rivers, countries, etc..).  This implmentation is more scheme like.
(defun-nondeterministic name0 (term)
  (let ((features (get term 'noun)))
    (if features
	(if (eq (cadr features) 
		 'name)
	    t 
	    (fail))
	(fail))))

;; ***  interrogatives are to be implemented later..
;(define int-art ())

;(define (int-pron))	  



(defun-nondeterministic adverb (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (adverb0 term)
	 (returned-values (list 'adv term) rem stack)))

(defun-nondeterministic adverb0 (term) 
  (let ((features (get term 'adverb)))
    (if features
	t
	(fail))))



(defun-nondeterministic poss-pron (phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-poss-pron gender agmt)
		  (poss-pron0 term)))
		(returned-values parsed-poss-pron gender agmt rem stack))))

(defun-nondeterministic poss-pron0 (term)
  (let ((features (get term 'poss-pron)))
    (if (not features)
	(fail)
	(returned-values (car features) (second features) (third features)))))


(defun-nondeterministic pers-pron (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-pron gender agmt case) 
		  (pers-pron0 term)))
		(returned-values parsed-pron gender agmt  
			case rem stack))))

(defun-nondeterministic pers-pron0 (term)
  (let ((features (get term 'pers-pron)))
    (if features
	(returned-values (car features) (second features) (third features) 
		(fourth features))
	(fail))))

(defun-nondeterministic compound-pers-pron (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((parsed-pron gender agmt case) 
		  (compound-pers-pron0 term)))
		(returned-values parsed-pron gender agmt  
			case rem stack))))

(defun-nondeterministic compound-pers-pron0 (term)
  (let ((features (get term 'comp-pers-pron)))
    (if features
	(returned-values (car features) (second features) (third features) 
		(fourth features))
	(fail))))

(defun-nondeterministic quantifier-pron (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((det-val noun-val) 
		  (quantifier-pron0 term)))
		(returned-values det-val noun-val rem stack))))

(defun-nondeterministic quantifier-pron0 (term)
 (let ((features (get term 'quantifier-pron)))
    (if (not features)
	(fail)
	(returned-values (car features) (cdr features)))))



(defun-nondeterministic context-pron (phrase xlist-stack)
  (either (mvlet (((term rem stack) 
		   (terminal phrase xlist-stack)))
		 (if (eq term 'where)
		     (returned-values (list 'in) 'place rem stack)
		     (fail)))
	  (mvlet (((term rem stack)
		   (terminal phrase xlist-stack)))
		 (if (eq term 'when)
		     (returned-values (list 'at) 'time rem stack)
		     (fail)))))

;; this should be removed	  
(defun-nondeterministic number (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (mvlet (((I number-val)
		  (number0 term)))
		(returned-values (list 'nb I) number-val rem stack))))

(defun-nondeterministic number0 (num)
  (fail))


(defun-nondeterministic terminator (phrase xlist-stack)
  (mvlet (((term rem stack) 
	   (terminal phrase xlist-stack)))
	 (returned-values  (terminator0 term) rem stack)))

(defun-nondeterministic terminator0 (term)
  (let ((type (get term 'terminator)))
    (if (not type)
	type
	(fail))))
		       
(defun-nondeterministic opt-the (phrase xlist-stack)
  (either (returned-values phrase xlist-stack)
	  (mvlet (((term rem stack)
		   (terminal phrase xlist-stack)))
		 (if (eq term 'the)
		     (returned-values rem stack)
		     (fail)))))


(defun-nondeterministic conj-val (phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (if (eq (get term 'synt-cat)
		 'conj)
	     (returned-values (list 'conj term) 'list 'end rem stack)
	     (fail))))


(defun-nondeterministic loc-pred (phrase xlist-stack)
  (mvlet (((term rem stack)
	   (terminal phrase xlist-stack)))
	 (returned-values (loc-pred0 term) rem stack)))

(defun-nondeterministic loc-pred0 (term)
  (let ((prep-equiv (get term 'loc-pred)))
    (if (not prep-equiv)
	prep-equiv
	(fail))))




