;;; -*- Mode: LISP; Package: SCREAMER; Base: 10; Syntax: Common-Lisp -*-

;;; LaHaShem HaAretz U'Mloah

(in-package 'screamer)

(export '(undefun-nondeterministic
          forward-declare-nondeterministic
          defun-nondeterministic
          either
          fail
          locally-setf
          one-value
          all-values
          one-of
          funcall-nondeterministic
          apply-nondeterministic
          lambda-nondeterministic))

;;; TTMTTD
;;;  1. Document
;;;  2. Problem Set #2

;;; Limitations
;;;  1. Overly stringent restriction on forms.
;;;  2. Macro expansion doesn't use local environment.
;;;  3. Could optimize CPS conversion of special forms if no subforms are
;;;     nondeterministic.
;;;  4. There are some possible optimizations to LOCALLY-SETF, PROGN,
;;;     SETQ and IF when you know that the result is not needed and the form is
;;;     only being evaluated for side effects.
;;;  5. Default forms for &OPTIONAL, &KEY and &AUX arguments to
;;;     DEFUN-NONDETERMINISTIC are not nondeterministic.
;;;  6. Do not handle SETF methods with multiple values for LOCALLY-SETF
;;;  7. Don't handle most Common Lisp special forms.
;;;     Currently handle:
;;;       QUOTE
;;;       FUNCTION
;;;       PROGN
;;;       SETQ
;;;       IF
;;;       LET
;;;       LET*
;;;     Maybe will someday handle:
;;;       CATCH
;;;       THROW
;;;       BLOCK
;;;       RETURN-FROM
;;;       TAGBODY
;;;       GO
;;;       UNWIND-PROTECT
;;;       MULTIPLE-VALUE-CALL
;;;       MULTIPLE-VALUE-PROG1
;;;     Probably will never handle:
;;;       COMPILER-LET
;;;       DECLARE
;;;       EVAL-WHEN
;;;       FLET
;;;       GENERIC-FLET
;;;       GENERIC-LABELS
;;;       LABELS
;;;       LOAD-TIME-VALUE
;;;       LOCALLY
;;;       MACROLET
;;;       PROGV
;;;       SYMBOL-MACROLET
;;;       THE
;;;       WITH-ADDED-METHODS

(eval-when (compile load eval)

(setf *NONDETERMINISM-FLAG* '*nondeterminism-flag*)

;; "The list of all nondeterministic functions"
(setf *NONDETERMINISTIC-FUNCTIONS* '())

;;  "The trail"
(setf *TRAIL* (make-array 0 :adjustable t :fill-pointer 0))

(defmacro CHOICE-POINT (form)
  `(let ((trail-pointer (fill-pointer *trail*)))
     (catch 'fail
       (unwind-protect ,form
         (loop (when (= (fill-pointer *trail*) trail-pointer) (return))
               (funcall (vector-pop *trail*)))))))

(defun IS-MAGIC-DECLARATION? (form)
  (equal form '(declare (#+symbolics sys:downward-function
                         #-symbolics ignore))))

(defun POSSIBLY-BETA-CONVERT-FUNCALL (continuation form)
  (if (and (consp continuation)
           (eq (first continuation) 'function)
           (eql (list-length continuation) 2)
           (consp (second continuation))
           (eq (first (second continuation)) 'lambda)
           (not (null (list-length (second continuation))))
           (>= (list-length (second continuation)) 3)
           (consp (second (second continuation)))
           (eql (list-length (second (second continuation))) 1)
           (symbolp (first (second (second continuation))))
           (is-magic-declaration? (third (second continuation))))
      `(let ((,(first (second (second continuation))) ,form))
         ,@(rest (rest (rest (second continuation)))))
      ;; note: the subst version of beta conversion is only valid when
      ;;       a. it gets substituted only once
      ;;       b. what gets substituted in has no side effects
      ;;       since I don't know how to check the later, for now the subst
      ;;       form of beta conversion has to be eliminated
      ;;(if (null (symbol-package (first (second (second continuation)))))
      ;;    (subst form
      ;;           (first (second (second continuation)))
      ;;           (case (list-length (second continuation))
      ;;             (3 ''nil)
      ;;             (4 (fourth (second continuation)))
      ;;             (otherwise
      ;;              `(progn ,@(rest (rest (rest (second continuation)))))))
      ;;           :test #'eq)
      ;;    `(let ((,(first (second (second continuation))) ,form))
      ;;       ,@(rest (rest (rest (second continuation))))))
      `(funcall ,continuation ,form)))

(defun CPS-CONVERT-FUNCTION-NAME (function-name)
  (intern (format nil "~A-NONDETERMINISTIC" (string function-name))
          (symbol-package function-name)))

(defun CPS-CONVERT-EITHER (forms continuation)
  (case (length forms)
    (0 `(throw 'fail nil))
    (1 (cps-convert (first forms) continuation))
    (otherwise
     (if (symbolp continuation)
         `(progn
            ,@(mapcar #'(lambda (form)
                          `(choice-point ,(cps-convert form continuation)))
                      (reverse (rest (reverse forms))))
            ,(cps-convert (first (last forms)) continuation))
         (let ((c (gensym "CONTINUATION-")))
           `(let ((,c ,continuation))
              ,@(mapcar #'(lambda (form) `(choice-point ,(cps-convert form c)))
                        (reverse (rest (reverse forms))))
              ,(cps-convert (first (last forms)) c)))))))

(defun CPS-CONVERT-LOCALLY-SETF (arguments continuation)
  ;; note: Could optimize away LOCALLY-SETF CPS conversion if no
  ;;       subform of LOCALLY-SETF is nondeterministic.
  (if (null arguments)
      (possibly-beta-convert-funcall continuation 'nil)
      (let ((d (gensym "DUMMY-")))
        (multiple-value-bind (vars vals stores store-form access-form)
            (get-setf-method (first arguments))
          (cps-convert
           (second arguments)
           `#'(lambda (,(first stores))
                (declare (#+symbolics sys:downward-function
                          #-symbolics ignore))
                (let* ,(mapcar #'list
                               (append vars (list d))
                               (append vals (list access-form)))
                  (vector-push-extend
                   #'(lambda () ,(subst d (first stores) store-form))
                   *trail*)
                  ,store-form
                  ,(if (null (rest (rest arguments)))
                       (possibly-beta-convert-funcall
                        continuation (first stores))
                       (cps-convert-locally-setf
                        (rest (rest arguments)) continuation)))))))))

(defun CPS-CONVERT-CALL (function-name arguments dummy-arguments continuation)
  (if (null arguments)
      `(,(cps-convert-function-name function-name)
        ,continuation
        ,@(reverse dummy-arguments))
      (let ((dummy-argument (gensym "DUMMY-")))
        (cps-convert
         (first arguments)
         `#'(lambda (,dummy-argument)
              (declare (#+symbolics sys:downward-function
                        #-symbolics ignore))
              ,(cps-convert-call
                function-name
                (rest arguments)
                (cons dummy-argument dummy-arguments)
                continuation))))))

(defun CPS-NON-CONVERT-CALL
       (function-name arguments dummy-arguments continuation)
  (if (null arguments)
      (possibly-beta-convert-funcall
       continuation `(,function-name ,@(reverse dummy-arguments)))
      (let ((dummy-argument (gensym "DUMMY-")))
        (cps-convert
         (first arguments)
         `#'(lambda (,dummy-argument)
              (declare (#+symbolics sys:downward-function
                        #-symbolics ignore))
              ,(cps-non-convert-call
                function-name
                (rest arguments)
                (cons dummy-argument dummy-arguments)
                continuation))))))

(defun CPS-CONVERT-PROGN (arguments continuation)
  ;; note: Could optimize away PROGN CPS conversion if no subform of PROGN
  ;;       is nondeterministic.
  (cond ((null arguments)
         (possibly-beta-convert-funcall continuation 'nil))
        ((null (rest arguments))
         (cps-convert (first arguments) continuation))
        (t (let ((dummy-argument (gensym "DUMMY-")))
             (cps-convert
              (first arguments)
              `#'(lambda (,dummy-argument)
                   (declare (#+symbolics sys:downward-function
                             #-symbolics ignore))
                   ,dummy-argument
                   ,(cps-convert-progn (rest arguments) continuation)))))))

(defun CPS-CONVERT-SETQ (arguments continuation)
  ;; note: Could optimize away SETQ CPS conversion if no subform of SETQ
  ;;       is nondeterministic.
  (if (null arguments)
      (possibly-beta-convert-funcall continuation 'nil)
      (let ((dummy-argument (gensym "DUMMY-")))
        (cps-convert
         (second arguments)
         `#'(lambda (,dummy-argument)
              (declare (#+symbolics sys:downward-function
                        #-symbolics ignore))
              (setq ,(first arguments) ,dummy-argument)
              ,(if (null (rest (rest arguments)))
                   (possibly-beta-convert-funcall continuation dummy-argument)
                   (cps-convert-setq (rest (rest arguments)) continuation)))))))

(defun CPS-CONVERT-IF (antecedent consequent alternate continuation)
  ;; note: Could optimize away IF CPS conversion if no subform of IF
  ;;       is nondeterministic.
  (let ((dummy-argument (gensym "DUMMY-"))
        (c (gensym "CONTINUATION-")))
    (if (symbolp continuation)
        (cps-convert
         antecedent
         `#'(lambda (,dummy-argument)
              (declare (#+symbolics sys:downward-function
                        #-symbolics ignore))
              (if ,dummy-argument
                  ,(cps-convert consequent continuation)
                  ,(cps-convert alternate continuation))))
        `(let ((,c ,continuation))
           ,(cps-convert
             antecedent
             `#'(lambda (,dummy-argument)
                  (declare (#+symbolics sys:downward-function
                            #-symbolics ignore))
                  (if ,dummy-argument
                      ,(cps-convert consequent c)
                      ,(cps-convert alternate c))))))))

(defun CPS-CONVERT-LET (bindings body continuation &optional new-bindings)
  (if (null bindings)
      `(let ,new-bindings ,(cps-convert body continuation))
      (let* ((binding (first bindings))
             (binding-variable
              (if (symbolp binding) binding (first binding)))
             (binding-form
              (if (and (consp binding) (eql (list-length binding) 2))
                  (second binding)
                  ''nil))
             (dummy-argument (gensym "DUMMY-")))
        (cps-convert
         binding-form
         `#'(lambda (,dummy-argument)
              (declare (#+symbolics sys:downward-function
                        #-symbolics ignore))
              ,(cps-convert-let (rest bindings)
                                body
                                continuation
                                (cons (list binding-variable dummy-argument)
                                      new-bindings)))))))

(defun CPS-CONVERT-LET* (bindings body continuation)
  (if (null bindings)
      (cps-convert body continuation)
      (let* ((binding (first bindings))
             (binding-variable
              (if (symbolp binding) binding (first binding)))
             (binding-form
              (if (and (consp binding) (eql (list-length binding) 2))
                  (second binding)
                  ''nil)))
        (cps-convert
         binding-form
         `#'(lambda (,binding-variable)
              (declare (#+symbolics sys:downward-function
                        #-symbolics ignore))
              ,(cps-convert-let* (rest bindings) body continuation))))))

(defun CPS-CONVERT-LAMBDA-NONDETERMINISTIC (lambda-list body)
  (let ((continuation (gensym "CONTINUATION-"))
	(flag (gensym "FLAG-")))
    `(lambda (,flag ,continuation ,@lambda-list)
       (if (eq ,flag *nondeterminism-flag*)
           ,(cps-convert-progn body continuation)
           (error "You attempted to call a LAMBDA-NONDETERMINISTIC~%~
                   expression either directly or indirectly via a~%~
                   deterministic FUNCALL or APPLY. You must use~%~
                   FUNCALL-NONDETERMINISTIC or APPLY-NONDETERMINISTIC~%~
                   instead.")))))

(defun CPS-CONVERT (form continuation)
  (if (consp form)
      (cond
       ;; note: These three restrictions are safe but somewhat too stringent.
       ((null (list-length form)) (error "Form ~A is circular" form))
       ((not (null (cdr (last form)))) (error "Form ~A is improper" form))
       ((not (symbolp (first form)))
        (cond ((and (listp (first form))
                    (not (null (list-length (first form))))
                    (>= (list-length (first form)) 2)
                    (eq (first (first form)) 'lambda)
                    (listp (second (first form))))
               (if (not (every #'(lambda (argument)
                                   (and (symbolp argument)
                                        (not (member argument
                                                     lambda-list-keywords
                                                     :test #'eq))))
                               (second (first form))))
                   (error "Cannot (currently) handle a form ~A whose CAR is~%~
                           a LAMBDA expression with lambda list~%~
                           keywords or arguments that are not symbols"
                          form))
               (if (/= (length (second (first form))) (length (rest form)))
                   (error "The form ~A has a CAR which is a LAMBDA~%~
                           expression which takes a different number of~%~
                           arguments than it is called with"
                          form))
               (cps-convert-let
                (mapcar #'list (second (first form)) (rest form))
                (cond
                 ((null (rest (rest (first form)))) ''nil)
                 ((null (rest (rest (rest (first form))))) (third (first form)))
                 (t `(progn ,@(rest (rest (first form))))))
                continuation))
              (t (error "CAR of form ~A is neither a symbol nor a~%~
                         valid LAMBDA expression" form))))
       ((eq (first form) 'either) (cps-convert-either (rest form) continuation))
       ((eq (first form) 'fail)
        (if (not (eql (list-length form) 1))
            (error "FAIL does not take any arguments: ~A" form))
        (cps-convert-either '() continuation))
       ((eq (first form) 'locally-setf)
        (if (oddp (length (rest form)))
            (error "Odd number of arguments to LOCALLY-SETF: ~A" form))
        (cps-convert-locally-setf (rest form) continuation))
       ;; note: Macro, expansion will not use the local environment so things
       ;;       like calls to macros defined with MACROLET will fail.
       ((macro-function (first form))
        (cps-convert (macroexpand form) continuation))
       ((special-form-p (first form))
        (case (first form)
          (quote
           (if (not (eql (list-length form) 2))
               (error "QUOTE must take one argument: ~A" form))
           (possibly-beta-convert-funcall continuation form))
          (function
           (if (not (eql (list-length form) 2))
               (error "FUNCTION must take one argument: ~A" form))
           (if (and (consp (second form))
                    (eq (first (second form)) 'lambda-nondeterministic))
               (if (and (not (null (list-length (second form))))
                        (> (list-length (second form)) 2)
                        (consp (second (second form))))
                   (possibly-beta-convert-funcall
                    continuation
                    `#',(cps-convert-lambda-nondeterministic
                         (second (second form))
                         (rest (rest (second form)))))
                   (error "Invalid syntax for LAMBDA-NONDETERMINISTIC: ~A"
                          form))
               (possibly-beta-convert-funcall continuation form)))
          (progn (cps-convert-progn (rest form) continuation))
          (setq
           (if (oddp (length (rest form)))
               (error "Odd number of arguments to SETQ: ~A" form))
           (cps-convert-setq (rest form) continuation))
          (if
           (if (not (or (eql (list-length form) 3) (eql (list-length form) 4)))
               (error "IF must take two or three arguments: ~A" form))
           (cps-convert-if (second form)
                           (third form)
                           (if (null (rest (rest (rest form))))
                               'nil
                               (fourth form))
                           continuation))
          (let
           (if (null (rest form))
            (error "LET must have BINDINGS: ~A" form))
            (if (or (not (listp (second form)))
                    (some #'(lambda (binding)
                              (not (or (symbolp binding)
                                       (and (consp binding)
                                            (or (eql (list-length binding) 1)
                                                (eql (list-length binding) 2))
                                            (symbolp (first binding))))))
                          (second form)))
                (error "Invalid BINDINGS for LET: ~A" form))
            (cps-convert-let
             (second form)
             (cond ((null (rest (rest form))) ''nil)
                   ((null (rest (rest (rest form)))) (third form))
                   (t `(progn ,@(rest (rest form)))))
             continuation))
          (let*
           (if (null (rest form))
            (error "LET* must have BINDINGS: ~A" form))
            (if (or (not (listp (second form)))
                    (some #'(lambda (binding)
                              (not (or (symbolp binding)
                                       (and (consp binding)
                                            (or (eql (list-length binding) 1)
                                                (eql (list-length binding) 2))
                                            (symbolp (first binding))))))
                          (second form)))
                (error "Invalid BINDINGS for LET*: ~A" form))
            (cps-convert-let*
             (second form)
             (cond ((null (rest (rest form))) ''nil)
                   ((null (rest (rest (rest form)))) (third form))
                   (t `(progn ,@(rest (rest form)))))
             continuation))
          (otherwise
           (error "Cannot (currently) handle the special form ~A inside a~%~
                   nondeterministic context. There are three types of~%~
                   nondeterministic contexts: The body of a function defined~%~
                   with DEFUN-NONDETERMINISTIC, the body of a call to the~%~
                   ALL-VALUES macro, and the first argument of a call to the~%~
                   ONE-VALUE macro. In particular, the body of a LAMBDA~%~
                   expression nested in a nondeterministic is a~%~
                   deterministic context. Likewise, the default forms of~%~
                   &OPTIONAL and &KEY arguments and the initialization forms~%~
                   of &AUX variables, are always deterministic contexts even~%~
                   though they may appear in DEFUN-NONDETERMINISTIC."
                  (first form)))))
       ((member (first form) *nondeterministic-functions* :test #'eq)
        (cps-convert-call (first form) (rest form) '() continuation))
       (t (cps-non-convert-call
           (first form) (rest form) '() continuation)))
      (possibly-beta-convert-funcall continuation form)))

(defun LAMBDA-LIST-VARIABLES (lambda-list)
  ;; note: this kludge may fail when lambda-list or its sublists are not
  ;;       proper, non-circular lists
  (set-difference
   (reduce #'union
           (mapcar #'(lambda (arg)
                       (if (consp arg)
                           (if (eql (list-length arg) 3)
                               (list (if (and (consp (first arg))
                                              (eql (list-length arg) 2))
                                         (second (first arg))
                                         (first arg))
                                     (third arg))
                               (list (if (and (consp (first arg))
                                              (eql (list-length arg) 2))
                                         (second (first arg))
                                         (first arg))))
                           (list arg)))
                   lambda-list)
           :initial-value '())
   lambda-list-keywords
   :test #'eq))

(defun DECLARE-NONDETERMINISTIC (function-name)
  (if (not (symbolp function-name))
      (error "Function name ~A must be a symbol" function-name))
  (setf *nondeterministic-functions*
        (adjoin function-name *nondeterministic-functions* :test #'eq)))

;;; note: The deterministic version of a nondeterministic function
;;;       must be capable of behaving as the nondeterministic function
;;;       itself when passed as a functional argument to procedures like
;;;       funcall-nondeterminisitic.  In most cases, however, the determinisitic
;;;       version should return an error.

;;; note: c-sh-A on a a function defined with DEFUN-NONDETERMINISTIC
;;;       return the wrong results since it will have a flag as the first
;;;       argument and a continuation as the second argument.

(defun DETERMINISTIC-DEFINITION (function-name continuation lambda-list body)
  (let ((flag (gensym "FLAG-")))
    `(defun ,function-name (,flag ,continuation ,@lambda-list)
       (if (eq ,flag *nondeterminism-flag*)
           ,(cps-convert-progn body continuation)
	   (error "Function ~A is a nondeterministic function. As such, it~%~
                   must be called only from a nondeterministic context.~%~
                   There are three types of nondeterministic contexts:~%~
                   The body of a function defined with~%~
                   DEFUN-NONDETERMINISTIC, the body of a call to the~%~
                   ALL-VALUES macro, and the first argument of a call to the~%~
                   ONE-VALUE macro. In particular, the body of a LAMBDA~%~
                   expression nested in a nondeterministic is a~%~
                   deterministic  context. Likewise, the default forms of~%~
                   &OPTIONAL and &KEY arguments and the initialization forms~%~
                   of &AUX variables, are always deterministic contexts even~%~
                   though they may appear in DEFUN-NONDETERMINISTIC. A~%~
                   nondeterministic cannot be FUNCALLed or APPLYed. Use~%~
                   FUNCALL-NONDETERMINISTIC or APPLY-NONDETERMINISTIC~%~
                   instead. Perhaps you forgot to declare ~A to be~%~
                   nondeterministic via a FORWARD-DECLARE-NONDETERMINISTIC~%~
                   declaration."
		  ',function-name
		  ',function-name))))))

;;; The protocol

(defun UNDEFUN-NONDETERMINISTIC (function-name)
  (cond
   ((member function-name *nondeterministic-functions* :test #'eq)
    (setf *nondeterministic-functions*
          (delete function-name *nondeterministic-functions* :test #'eq))
    (fmakunbound function-name)
    (fmakunbound (cps-convert-function-name function-name)))
   (t (error "There is no nondeterministic function named ~A" function-name))))

(defmacro FORWARD-DECLARE-NONDETERMINISTIC (function-name)
  `(eval-when (compile load eval)
     (declare-nondeterministic ',function-name)))

(defmacro DEFUN-NONDETERMINISTIC (function-name lambda-list &body body)
  (declare-nondeterministic function-name)
  (let ((continuation (gensym "CONTINUATION-")))
    `(eval-when (compile load eval)
       (declare-nondeterministic ',function-name)
       ,(deterministic-definition function-name continuation lambda-list body)
       (defun ,(cps-convert-function-name function-name)
              (,continuation ,@lambda-list)
         ,(cps-convert-progn body continuation)))))

(defmacro EITHER (&body forms)
  (case (length forms)
    (0 `(throw 'fail nil))
    (1 (first forms))
    (otherwise
     (error "You can only use the EITHER special form in a nondeterministic~%~
             context. There are three types of nondeterministic contexts:~%~
             The body of a function defined with DEFUN-NONDETERMINISTIC,~%~
             the body of a call to the ALL-VALUES macro, and the first~%~
             argument of a call to the ONE-VALUE macro. In particular, the~%~
             body of a LAMBDA expression nested in a nondeterministic~%~
             is a deterministic context. Likewise, the default forms of~%~
             &OPTIONAL and &KEY arguments and the initialization forms of~%~
             &AUX variables, are always deterministic contexts even though~%~
             they may appear in DEFUN-NONDETERMINISTIC. Also, EITHER is not~%~
             a function so it cannot be FUNCALLed or APPLYed."))))

;;; note: Needs work to signal an error if the following two macros are called
;;;       in a deterministic context which is not nested in a choice point.

(defmacro FAIL () `(throw 'fail nil))

(defmacro LOCALLY-SETF (reference value &rest more-pairs)
  (if (oddp (length more-pairs))
      (error "Odd number of arguments to LOCALLY-SETF: ~A"
             `(locally-setf ,reference ,value ,@more-pairs)))
  (let ((d (gensym "DUMMY-")))
    (multiple-value-bind (vars vals stores store-form access-form)
        (get-setf-method reference)
      `(let* ,(mapcar #'list
                      (append vars (list (first stores) d))
                      (append vals (list value access-form)))
         (vector-push-extend
          #'(lambda () ,(subst d (first stores) store-form))
          *trail*)
         ,@(if (null more-pairs)
               (list store-form)
               (list store-form `(locally-setf ,@more-pairs)))))))

;;; note: Ideally, in a nondeterministic context, a ONE-VALUE with one argument
;;;       that does not have any possible values should fail rather tha signal
;;;       an error.

(defmacro ONE-VALUE (form1 &optional (form2 nil form2?))
  `(catch 'succeed
     (choice-point
      ,(cps-convert
        form1 '#'(lambda (value)
                   (declare (#+symbolics sys:downward-function
                             #-symbolics ignore))
                   (throw 'succeed value))))
     ,(if form2? form2 `(error "There is no value for the form: ~A" ',form1))))

(defmacro ALL-VALUES (&body forms)
  `(let ((values '())
         (last-value-cons nil))
     (choice-point
      ,(cps-convert-progn
        forms
        '#'(lambda (value)
             (declare (#+symbolics sys:downward-function
                       #-symbolics ignore))
             (cond ((null values)
                    (setf last-value-cons (list value))
                    (setf values last-value-cons))
                   (t (setf (rest last-value-cons) (list value))
                      (setf last-value-cons (rest last-value-cons))))
             (throw 'fail nil))))
     values))

;;; note: Should have way of having a stream of values as well as a continuable
;;;       top level.

(defun-nondeterministic ONE-OF-VECTOR (vector i n)
  (if (= i n) (fail) (either (aref vector i) (one-of-vector vector (1+ i) n))))

(defun-nondeterministic ONE-OF (values)
  (cond
   ((listp values)
    (if (null values) (fail) (either (first values) (one-of (rest values)))))
   ((vectorp values) (one-of-vector values 0 (length values)))
   (t (error "VALUES must be a sequence"))))

;;; note: c-sh-A on FUNCALL-NONDETERMINISTIC and APPLY-NONDETERMINISTIC will
;;;       return the wrong results since it will have a flag as the first
;;;       argument and a continuation as the second argument.

(forward-declare-nondeterministic funcall-nondeterministic)

(defun FUNCALL-NONDETERMINISTIC (flag continuation function &rest arguments)
  (if (eq flag *nondeterminism-flag*)
      (apply function flag continuation arguments)
      (error "FUNCALL-NONDETERMINISTIC is a nondeterministic function. As~%~
              such, it must be called only from a nondeterministic context.~%~
              There are three types of nondeterministic contexts:~%~
              The body of a function defined with~%~
              DEFUN-NONDETERMINISTIC, the body of a call to the~%~
              ALL-VALUES macro, and the first argument of a call to the~%~
              ONE-VALUE macro. In particular, the body of a LAMBDA~%~
              expression nested in a nondeterministic is a~%~
              deterministic  context. Likewise, the default forms of~%~
              &OPTIONAL and &KEY arguments and the initialization forms~%~
              of &AUX variables, are always deterministic contexts even~%~
              though they may appear in DEFUN-NONDETERMINISTIC. A~%~
              nondeterministic cannot be FUNCALLed or APPLYed. Use~%~
              FUNCALL-NONDETERMINISTIC or APPLY-NONDETERMINISTIC~%~
              instead.")))

(defun FUNCALL-NONDETERMINISTIC-NONDETERMINISTIC
       (continuation function &rest arguments)
  (apply function *nondeterminism-flag* continuation arguments))

(forward-declare-nondeterministic apply-nondeterministic)

(defun APPLY-NONDETERMINISTIC
       (flag continuation function argument &rest arguments)
  (if (eq flag *nondeterminism-flag*)
      (apply function
             flag
             continuation
             ;; note: I don't know how to avoid the consing here.
             (apply #'list* (cons argument arguments)))
      (error "FUNCALL-NONDETERMINISTIC is a nondeterministic function. As~%~
              such, it must be called only from a nondeterministic context.~%~
              There are three types of nondeterministic contexts:~%~
              The body of a function defined with~%~
              DEFUN-NONDETERMINISTIC, the body of a call to the~%~
              ALL-VALUES macro, and the first argument of a call to the~%~
              ONE-VALUE macro. In particular, the body of a LAMBDA~%~
              expression nested in a nondeterministic is a~%~
              deterministic  context. Likewise, the default forms of~%~
              &OPTIONAL and &KEY arguments and the initialization forms~%~
              of &AUX variables, are always deterministic contexts even~%~
              though they may appear in DEFUN-NONDETERMINISTIC. A~%~
              nondeterministic cannot be FUNCALLed or APPLYed. Use~%~
              FUNCALL-NONDETERMINISTIC or APPLY-NONDETERMINISTIC~%~
              instead.")))

(defun APPLY-NONDETERMINISTIC-NONDETERMINISTIC
       (continuation function argument &rest arguments)
  (apply function
         *nondeterminism-flag*
         continuation
         ;; note: I don't know how to avoid the consing here.
         (apply #'list* (cons argument arguments))))

(defmacro LAMBDA-NONDETERMINISTIC (lambda-list &body body)
  (cps-convert-lambda-nondeterministic lambda-list body))

;;; Tam V'Nishlam Shevah L'El Borei Olam

